# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT


import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from MuyGPyS._test.gp import benchmark_sample, BenchmarkGP
from MuyGPyS.gp.distortion import IsotropicDistortion
from MuyGPyS.gp.hyperparameter import ScalarHyperparameter
from MuyGPyS.gp.kernels import Matern
from MuyGPyS.gp.noise import HomoscedasticNoise
from MuyGPyS.optimize.loss import mse_fn


class SamplerBase:
    def __init__(self, data_count, train_ratio=0.1, seed=0):
        rng = np.random.default_rng(seed=seed)
        interval_count = int(data_count * train_ratio)
        interval = int(data_count / interval_count)
        sfl = rng.permutation(np.arange(data_count))
        self._train_mask = np.zeros(data_count, dtype=bool)
        for i in range(interval_count):
            idx = np.random.choice(sfl[i * interval : (i + 1) * interval])
            self._train_mask[idx] = True
        self._test_mask = np.invert(self._train_mask)
        self._title_fsize = 18
        self._axis_fsize = 14
        self._legend_fsize = 12


class UnivariateSampler(SamplerBase):
    def __init__(
        self,
        data_count=5001,
        train_ratio=0.1,
        seed=0,
        view_lb=0.5,
        view_ub=0.6,
        kernel=Matern(
            nu=ScalarHyperparameter(2.0),
            metric=IsotropicDistortion(
                "l2", length_scale=ScalarHyperparameter(1.0)
            ),
        ),
        eps=HomoscedasticNoise(1e-14),
        measurement_eps=HomoscedasticNoise(1e-5),
    ):
        super().__init__(data_count, train_ratio=train_ratio, seed=seed)
        self.data_count = data_count
        self.x = np.linspace(0, 1, data_count).reshape(data_count, 1)
        self.test_features = self.x[self._test_mask, :]
        self.train_features = self.x[self._train_mask, :]
        self.test_count, _ = self.test_features.shape
        self.train_count, _ = self.train_features.shape
        self.measurement_eps = measurement_eps
        self.gp = BenchmarkGP(kernel=kernel, eps=eps)
        self.train_interval = self.get_interval(
            view_lb, view_ub, self.train_features
        )
        self.test_interval = self.get_interval(
            view_lb, view_ub, self.test_features
        )
        self._target_color = "#7570b3"
        self._predict_colors = ["#d95f02", "#1b9e77"]

    def features(self):
        return self.train_features, self.test_features

    def sample(self):
        y = benchmark_sample(self.gp, self.x)
        self.test_responses = y[self._test_mask, :]
        self.train_responses = y[self._train_mask, :] + np.random.normal(
            0, self.measurement_eps(), size=(self.train_count, 1)
        )
        return self.train_responses, self.test_responses

    def plot_sample(self):
        _, axes = plt.subplots(2, 1, figsize=(8, 7))

        axes[0].set_title("Sampled Curve", fontsize=self._title_fsize)
        axes[0].set_xlabel("Feature Domain", fontsize=self._axis_fsize)
        axes[0].set_ylabel("Response Range", fontsize=self._axis_fsize)
        axes[0].plot(
            self.train_features,
            self.train_responses,
            "k*",
            label="perturbed train response",
        )
        axes[0].plot(
            self.test_features,
            self.test_responses,
            "-",
            color=self._target_color,
            label="test response",
        )
        axes[0].legend(fontsize=self._legend_fsize)

        self.plot_target_sub(axes[1])

        plt.tight_layout()

        plt.show()

    def plot_results(
        self,
        *args,
    ):
        _, axes = plt.subplots(2, 1, figsize=(8, 8))

        self.plot_target(axes[0])
        self.plot_target_sub(axes[1])
        for i, (name, predictions, confidence_intervals) in enumerate(args):
            # if i == 0:
            self.plot_model(
                axes[0],
                name,
                predictions,
                confidence_intervals,
                color=self._predict_colors[i],
            )

            self.plot_model_sub(
                axes[1],
                name,
                predictions,
                confidence_intervals,
                color=self._predict_colors[i],
            )

        plt.tight_layout()

        plt.show()

    def plot_target(self, ax):
        ax.set_title("Sampled Curve", fontsize=self._title_fsize)
        ax.set_xlabel("Feature Domain", fontsize=self._axis_fsize)
        ax.set_ylabel("Response Range", fontsize=self._axis_fsize)
        ax.plot(
            self.train_features,
            self.train_responses,
            "k*",
            label="perturbed train response",
        )
        ax.plot(
            self.test_features,
            self.test_responses,
            "-",
            color=self._target_color,
            label="test response",
        )

    def plot_model(
        self,
        ax,
        name,
        predictions,
        confidence_intervals,
        color=None,
    ):
        if color is None:
            color = self._predict_colors[0]
        confidence_intervals = confidence_intervals.reshape(self.test_count)
        ax.set_title(
            f"Sampled Curve with {name} model",
            fontsize=self._title_fsize,
        )
        ax.plot(
            self.test_features,
            predictions,
            "--",
            color=color,
            label=f"{name} predictions",
        )
        ax.fill_between(
            self.test_features[:, 0],
            (predictions[:, 0] - confidence_intervals),
            (predictions[:, 0] + confidence_intervals),
            facecolor=color,
            alpha=0.25,
            label=f"{name} 95% Confidence Interval",
        )
        ax.legend(fontsize=self._legend_fsize)

    def get_interval(self, lb, ub, ys):
        return np.array(
            [i for i, x in enumerate(ys) if x >= lb and x <= ub], dtype=int
        )

    def plot_target_sub(self, ax):
        ax.set_title("Sampled Curve (subset)", fontsize=20)
        ax.set_xlabel("Feature Domain", fontsize=16)
        ax.set_ylabel("Response Range", fontsize=16)
        ax.plot(
            self.train_features[self.train_interval],
            self.train_responses[self.train_interval],
            "k*",
            label="perturbed train response",
        )
        ax.plot(
            self.test_features[self.test_interval],
            self.test_responses[self.test_interval],
            "-",
            color=self._target_color,
            label="test response",
        )

    def plot_model_sub(
        self, ax, name, predictions, confidence_intervals, color=None
    ):
        if color is None:
            color = self._predict_colors[0]
        confidence_intervals = confidence_intervals.reshape(self.test_count)
        ax.set_title(
            f"Sampled Curve (subset) with {name} model",
            fontsize=self._title_fsize,
        )
        ax.plot(
            self.test_features[self.test_interval],
            predictions[self.test_interval],
            "--",
            color=color,
            label=f"{name} predictions",
        )
        ax.fill_between(
            self.test_features[self.test_interval][:, 0],
            (predictions[:, 0] - confidence_intervals)[self.test_interval],
            (predictions[:, 0] + confidence_intervals)[self.test_interval],
            facecolor=color,
            alpha=0.25,
            label=f"{name} 95% Confidence Interval",
        )
        ax.legend(fontsize=self._legend_fsize)


class UnivariateSampler2D(SamplerBase):
    def __init__(
        self,
        points_per_dim=60,
        train_ratio=10,
        kernel=Matern(
            nu=ScalarHyperparameter(2.0),
            metric=IsotropicDistortion(
                "l2", length_scale=ScalarHyperparameter(1.0)
            ),
        ),
        eps=HomoscedasticNoise(1e-14),
        measurement_eps=HomoscedasticNoise(1e-5),
    ):
        self.points_per_dim = points_per_dim
        self.data_count = self.points_per_dim**2
        super().__init__(self.data_count, train_ratio)
        x = np.linspace(0, 1, points_per_dim)
        xx, yy = np.meshgrid(x, x)
        self.xs = np.array(
            [
                [xx[i, j], yy[i, j]]
                for i in range(points_per_dim)
                for j in range(points_per_dim)
            ]
        )
        self.test_features = self.xs[self._test_mask, :]
        self.train_features = self.xs[self._train_mask, :]
        self.test_count, _ = self.test_features.shape
        self.train_count, _ = self.train_features.shape
        self.measurement_eps = measurement_eps
        self.gp = BenchmarkGP(kernel=kernel, eps=eps)

    def features(self):
        return self.train_features, self.test_features

    def sample(self):
        self.ys = benchmark_sample(self.gp, self.xs)
        self.test_responses = self.ys[self._test_mask, :]
        self.train_responses = self.ys[self._train_mask, :] + np.random.normal(
            0, self.measurement_eps(), size=(self.train_count, 1)
        )
        return self.train_responses, self.test_responses

    def plot_sample(self):
        fig, axes = plt.subplots(1, 3, figsize=(19, 4))

        vmin = np.nanmin(self.ys)
        vmax = np.nanmax(self.ys)

        self._label_ax(axes[0], "Sampled Surface")
        im0 = axes[0].imshow(
            self._make_im(self.ys, mode="all"), vmin=vmin, vmax=vmax
        )

        self._label_ax(axes[1], "Training Points")
        axes[1].imshow(
            self._make_im(self.train_responses, mode="train"),
            vmin=vmin,
            vmax=vmax,
        )

        self._label_ax(axes[2], "Testing Points")
        axes[2].imshow(self._make_im(self.test_responses), vmin=vmin, vmax=vmax)
        fig.colorbar(im0, ax=axes.ravel().tolist())

        plt.show()

    def _make_im(self, array, mode="test", range=False, add_inf=True):
        im = np.zeros(self.data_count)
        if mode == "test":
            im[self._test_mask, None] = array
            if add_inf is True:
                im[self._train_mask] = -np.inf
        elif mode == "train":
            im[self._train_mask, None] = array
            if add_inf is True:
                im[self._test_mask] = -np.inf
        else:
            im[:, None] = array
        if range is False:
            return im.reshape(self.points_per_dim, self.points_per_dim)
        else:
            vmin = np.nanmin(array)
            vmax = np.nanmax(array)
            return (
                im.reshape(self.points_per_dim, self.points_per_dim),
                vmin,
                vmax,
            )

    def plot_predictions(self, *args):
        pred_count = len(args)
        fig, axes = plt.subplots(
            1, pred_count + 1, figsize=(4 * (pred_count + 1), 4)
        )

        test_im, vmin, vmax = self._make_im(self.test_responses, range=True)
        pred_ims = list()
        names = list()
        for name, predictions in args:
            pred_im, pvmin, pvmax = self._make_im(predictions, range=True)
            pred_ims.append(pred_im)
            names.append(name)
            vmin = np.min([vmin, pvmin])
            vmax = np.max([vmax, pvmax])

        self._label_ax(axes[0], "Expected Surface")
        im0 = axes[0].imshow(
            test_im,
            vmin=vmin,
            vmax=vmax,
            cmap="viridis",
        )
        fig.colorbar(im0, ax=axes[0])

        for i, pred_im in enumerate(pred_ims):
            self._label_ax(axes[i + 1], f"{names[i]} Surface")
            im1 = axes[i + 1].imshow(
                pred_im,
                vmin=vmin,
                vmax=vmax,
                cmap="viridis",
            )
            fig.colorbar(im1, ax=axes[i + 1])

        plt.tight_layout()
        plt.show()

    def _get_images(self, predictions, confidence_intervals):
        resl_im, resl_min, resl_max = self._make_im(
            self.test_responses - predictions, range=True
        )
        conf_im, _, conf_mag = self._make_im(confidence_intervals, range=True)
        covr_im, covr_min, covr_max = self._make_im(
            np.abs(self.test_responses - predictions) - confidence_intervals,
            range=True,
        )

        resl_mag = np.max([np.abs(resl_min), np.abs(resl_max)])
        covr_mag = np.max([np.abs(covr_min), np.abs(covr_max)])
        return resl_im, conf_im, covr_im, resl_mag, conf_mag, covr_mag

    def plot_errors(self, *args):
        row_count = len(args)

        fig, axes = plt.subplots(row_count, 3, figsize=(13, 4 * row_count))
        if row_count == 1:
            name, predictions, confidence_intervals = args[0]
            (
                resl_im,
                conf_im,
                covr_im,
                resl_mag,
                conf_mag,
                covr_mag,
            ) = self._get_images(predictions, confidence_intervals)
            self.plot_error(
                fig,
                axes,
                name,
                resl_im,
                conf_im,
                covr_im,
                resl_mag,
                conf_mag,
                covr_mag,
            )
        else:
            resl_ims = []
            conf_ims = []
            covr_ims = []
            resl_mag = 0.0
            conf_mag = 0.0
            covr_mag = 0.0
            for i, (_, predictions, confidence_intervals) in enumerate(args):
                (
                    resl_im,
                    conf_im,
                    covr_im,
                    resl_mag_,
                    conf_mag_,
                    covr_mag_,
                ) = self._get_images(predictions, confidence_intervals)
                resl_ims.append(resl_im)
                conf_ims.append(conf_im)
                covr_ims.append(covr_im)
                resl_mag = np.max([resl_mag, resl_mag_])
                conf_mag = np.max([conf_mag, conf_mag_])
                covr_mag = np.max([covr_mag, covr_mag_])
            for i, (name, _, _) in enumerate(args):
                self.plot_error(
                    fig,
                    axes[i, :],
                    name,
                    resl_ims[i],
                    conf_ims[i],
                    covr_ims[i],
                    resl_mag,
                    conf_mag,
                    covr_mag,
                )
        plt.tight_layout()
        plt.show()

    def _label_ax(self, ax, title):
        ax.set_title(title, fontsize=self._title_fsize)
        ax.set_xlabel("Axis 0", fontsize=self._axis_fsize)
        ax.set_ylabel("Axis 1", fontsize=self._axis_fsize)
        ax.set_xticks([])
        ax.set_yticks([])

    def plot_error(
        self,
        fig,
        axes,
        name,
        resl_im,
        conf_im,
        covr_im,
        resl_mag,
        conf_mag,
        covr_mag,
    ):
        self._label_ax(axes[0], f"{name} residual")
        im0 = axes[0].imshow(
            resl_im, vmin=-resl_mag, vmax=resl_mag, cmap="coolwarm"
        )
        fig.colorbar(im0, ax=axes[0])

        self._label_ax(axes[1], f"{name} CI Magnitude")
        im1 = axes[1].imshow(conf_im, vmin=0.0, vmax=conf_mag, cmap="inferno")
        fig.colorbar(im1, ax=axes[1])

        self._label_ax(axes[2], f"{name} |Residual| - CI")
        im2 = axes[2].imshow(
            covr_im, vmin=-covr_mag, vmax=covr_mag, cmap="coolwarm"
        )
        fig.colorbar(im2, ax=axes[2])


def get_length_scale(muygps):
    ls = muygps.kernel.distortion_fn.length_scale
    if isinstance(ls, dict):
        return np.array([ls[x]() for x in ls])
    else:
        return ls()


def print_results(targets, *args, **kwargs):
    table = list()
    for arg in args:
        (
            name,
            muygps,
            means,
            variances,
            confidence_intervals,
            coverage,
        ) = arg
        table.append(
            [
                name,
                muygps.kernel.nu(),
                get_length_scale(muygps),
                muygps.eps(),
                muygps.sigma_sq()[0],
                np.sqrt(mse_fn(means, targets)),
                np.mean(variances),
                np.mean(confidence_intervals),
                coverage,
            ]
        )
    return pd.DataFrame(
        table,
        columns=[
            "name",
            "nu",
            "length scale",
            "$\\varepsilon$",
            "$\\sigma^2$",
            "rmse",
            "mean variance",
            "mean confidence interval",
            "coverage",
        ],
    ).style.hide(axis="index")


def print_fast_results(targets, *args, **kwargs):
    table = list()
    for arg in args:
        (
            name,
            time,
            muygps,
            means,
        ) = arg
        table.append(
            [
                name,
                np.sqrt(mse_fn(means, targets)),
                time,
                muygps.eps(),
            ]
        )
    return pd.DataFrame(
        table,
        columns=[
            "name",
            "rmse",
            "timing results",
            "$\\varepsilon$",
        ],
    ).style.hide(axis="index")



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

"""
Hyperparameters and kernel functors

Defines the Matérn kernel functor (inheriting
:class:`~MuyGPyS.gp.kernels.kernel_fn.KernelFn`) that transform crosswise and
pairwise difference tensors into cross-covariance and covariance (or kernel)
tensors, respectively.

See the following example to initialize an :class:`MuyGPyS.gp.kernels.Matern`
object.

Example:
    >>> from MuyGPyS.gp.kernels import Matern
    >>> kern = Matern(
    ...     nu=ScalarHyperparameter("log_sample", (0.1, 2.5)),
    ...     metric=IsotropicDistortion(
    ...         l2,
    ...         length_scale=ScalarHyperparameter(1.0),
    ...     ),
    ... )

One uses a previously computed `pairwise_diffs` tensor (see
:func:`MuyGPyS.gp.tensors.pairwise_tensor`) to compute a kernel tensor whose
second two dimensions contain square kernel matrices. Similarly, one uses a
previously computed `crosswise_diffs` matrix (see
:func:`MuyGPyS.gp.tensor.crosswise_tensor`) to compute a cross-covariance
matrix. See the following example, which assumes that you have already
constructed the differenece tensors and kernel as shown above.

Example:
    >>> K = kern(pairwise_diffs)
    >>> Kcross = kern(crosswise_diffs)
"""

from typing import Callable, List, Tuple, Union

import MuyGPyS._src.math as mm
from MuyGPyS._src.gp.kernels import (
    _matern_05_fn,
    _matern_15_fn,
    _matern_25_fn,
    _matern_inf_fn,
    _matern_gen_fn,
)

from MuyGPyS._src.util import auto_str
from MuyGPyS.gp.distortion import (
    embed_with_distortion_model,
    AnisotropicDistortion,
    IsotropicDistortion,
    NullDistortion,
    l2,
)
from MuyGPyS.gp.hyperparameter import ScalarHyperparameter
from MuyGPyS.gp.kernels import (
    KernelFn,
)


def _set_matern_fn(nu: ScalarHyperparameter):
    if nu.fixed() is True:
        if nu() == 0.5:
            return _matern_05_fn
        elif nu() == 1.5:
            return _matern_15_fn
        elif nu() == 2.5:
            return _matern_25_fn
        elif nu() == mm.inf:
            return _matern_inf_fn
        else:
            return _matern_gen_fn

    return _matern_gen_fn


@auto_str
class Matern(KernelFn):
    """
    The Matérn kernel.

    The Màtern kernel includes a parameterized distortion model
    :math:`d_\\ell(\\cdot, \\cdot)` and an additional smoothness parameter
    :math:`\\nu>0`. :math:`\\nu` is proportional to the smoothness of the
    resulting function. As :math:`\\nu\\rightarrow\\infty`, the kernel becomes
    equivalent to the :class:`RBF` kernel. When :math:`\\nu = 1/2`, the Matérn
    kernel is identical to the absolute exponential kernel. Important
    intermediate values are :math:`\\nu=1.5` (once differentiable functions) and
    :math:`\\nu=2.5` (twice differentiable functions).

    The kernel is defined by

    .. math::
         k(x_i, x_j) =  \\frac{1}{\\Gamma(\\nu)2^{\\nu-1}}\\Bigg(
         \\frac{\\sqrt{2\\nu}}{l} d_\\ell(x_i , x_j )
         \\Bigg)^\\nu K_\\nu\\Bigg(
         \\frac{\\sqrt{2\\nu}}{l} d(x_i , x_j )\\Bigg),

    where :math:`K_{\\nu}(\\cdot)` is a modified Bessel function and
    :math:`\\Gamma(\\cdot)` is the gamma function.
    Typically, :math:`d(\\cdot,\\cdot)` is the Euclidean distance or
    :math:`\\ell_2` norm of the difference of the operands.

    Args:
        nu:
            A hyperparameter dict defining the length_scale parameter.
        metric:
            The distance function to be used. Includes length_scale
            hyperparameter information via the MuyGPyS.gp.distortion module.
    """

    def __init__(
        self,
        nu: ScalarHyperparameter = ScalarHyperparameter(0.5),
        metric: Union[
            AnisotropicDistortion, IsotropicDistortion, NullDistortion
        ] = IsotropicDistortion(l2, length_scale=ScalarHyperparameter(1.0)),
    ):
        super().__init__(metric=metric)
        self.nu = nu
        self._make()

    def _make(self):
        super()._make_base()
        self._hyperparameters["nu"] = self.nu
        self._kernel_fn = _set_matern_fn(self.nu)
        self._fn = embed_with_distortion_model(
            self._kernel_fn,
            self.distortion_fn,
            self.distortion_fn.length_scale,
        )

    def __call__(self, diffs):
        """
        Compute Matern kernels from distance tensor.

        Takes inspiration from
        [scikit-learn](https://github.com/scikit-learn/scikit-learn/blob/95119c13a/sklearn/gaussian_process/kernels.py#L1529)

        Args:
            diffs:
                A tensor of pairwise differences of shape
                `(data_count, nn_count, nn_count, feature_count)`. It is assumed
                that the vectors along the diagonals diffs[i, j, j, :] == 0.
        Returns:
            A cross-covariance matrix of shape `(data_count, nn_count)` or a
            tensor of shape `(data_count, nn_count, nn_count)` whose last two
            dimensions are kernel matrices.
        """
        return self._fn(diffs, nu=self.nu())

    def get_opt_params(
        self,
    ) -> Tuple[List[str], List[float], List[Tuple[float, float]]]:
        """
        Report lists of unfixed hyperparameter names, values, and bounds.

        Returns
        -------
            names:
                A list of unfixed hyperparameter names.
            params:
                A list of unfixed hyperparameter values.
            bounds:
                A list of unfixed hyperparameter bound tuples.
        """
        names, params, bounds = super().get_opt_params()
        self.nu.append_lists("nu", names, params, bounds)
        return names, params, bounds

    def get_opt_fn(self) -> Callable:
        """
        Return a kernel function with fixed parameters set.
        This function is designed for use with
        :func:`MuyGPyS.optimize.chassis.optimize_from_tensors()` and assumes
        that optimization parameters will be passed as keyword arguments.

        Returns:
            A function implementing the kernel where all fixed parameters are
            set. The function expects keyword arguments corresponding to current
            hyperparameter values for unfixed parameters.
        """
        return self._get_opt_fn(self._fn, self.distortion_fn, self.nu)

    @staticmethod
    def _get_opt_fn(
        matern_fn: Callable,
        distortion_fn: Union[
            AnisotropicDistortion, IsotropicDistortion, NullDistortion
        ],
        nu: ScalarHyperparameter,
    ) -> Callable:
        opt_fn = KernelFn._get_opt_fn(matern_fn, distortion_fn)
        opt_fn = nu.apply(opt_fn, "nu")
        return opt_fn



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT


import MuyGPyS._src.math.numpy as np
import MuyGPyS._src.math.torch as torch


def _kk_fn(
    exp_inv_scaled_sum_sq_diffs,
    sum_sq_diffs,
    prod_sq_diffs,
    sum_quad_diffs,
    a=1,
    length_scale=1,
):
    return (
        1
        / 4
        * (
            a
            * (
                8 * length_scale**2
                - 8 * length_scale * sum_sq_diffs
                + 2 * prod_sq_diffs
                + sum_quad_diffs
            )
            * exp_inv_scaled_sum_sq_diffs
            / length_scale**4
        )
    )


def _kg1_fn(
    exp_inv_scaled_sum_sq_diffs,
    diff_xy_quad_diffs,
    diff_yx_sq_diffs,
    a=1,
    length_scale=1,
):
    return (
        1
        / 4
        * (
            a
            * (6 * length_scale * diff_yx_sq_diffs + diff_xy_quad_diffs)
            * exp_inv_scaled_sum_sq_diffs
            / length_scale**4
        )
    )


def _kg2_fn(
    exp_inv_scaled_sum_sq_diffs, sum_sq_diffs, prod_diffs, a=1, length_scale=1
):
    return (
        1
        / 4
        * (
            2
            * a
            * prod_diffs
            * (-6 * length_scale + sum_sq_diffs)
            * exp_inv_scaled_sum_sq_diffs
            / length_scale**4
        )
    )


def _g1g1_fn(
    exp_inv_scaled_sum_sq_diffs,
    sum_sq_diffs,
    sum_quad_diffs,
    prod_sq_diffs,
    a=1,
    length_scale=1,
):
    return (
        1
        / 4
        * (
            a
            * (
                4 * length_scale**2
                - 4 * length_scale * sum_sq_diffs
                - 2 * prod_sq_diffs
                + sum_quad_diffs
            )
            * exp_inv_scaled_sum_sq_diffs
            / length_scale**4
        )
    )


def _g1g2_fn(
    exp_inv_scaled_sum_sq_diffs,
    diff_xy_sq_diffs,
    prod_diffs,
    a=1,
    length_scale=1,
):
    return (
        1
        / 4
        * (
            2
            * a
            * prod_diffs
            * diff_xy_sq_diffs
            * exp_inv_scaled_sum_sq_diffs
            / length_scale**4
        )
    )


def _g2g2_fn(
    exp_inv_scaled_sum_sq_diffs,
    sum_sq_diffs,
    prod_sq_diffs,
    a=1,
    length_scale=1,
):
    return (
        1
        / 4
        * (
            4
            * a
            * (length_scale**2 - length_scale * sum_sq_diffs + prod_sq_diffs)
            * exp_inv_scaled_sum_sq_diffs
            / length_scale**4
        )
    )


# compute the full covariance matrix
def _shear_fn(diffs, a=1, length_scale=1):
    shape = torch.array(diffs.shape[:-1], dtype=int)
    shape[-1] *= 3
    shape[-2] *= 3
    full_m = torch.zeros(tuple(shape))

    # compute intermediate difference tensors once here
    prod_diffs = torch.prod(diffs, axis=-1)
    sq_diffs = diffs**2
    quad_diffs = sq_diffs**2
    sum_sq_diffs = torch.sum(sq_diffs, axis=-1)
    prod_sq_diffs = torch.prod(sq_diffs, axis=-1)
    sum_quad_diffs = torch.sum(quad_diffs, axis=-1)
    diff_yx_sq_diffs = sq_diffs[..., 1] - sq_diffs[..., 0]
    diff_xy_sq_diffs = sq_diffs[..., 0] - sq_diffs[..., 1]
    diff_xy_quad_diffs = quad_diffs[..., 0] - quad_diffs[..., 1]
    exp_inv_scaled_sum_sq_diffs = torch.exp(-sum_sq_diffs / (2 * length_scale))

    full_m[..., 0::3, 0::3] = _kk_fn(
        exp_inv_scaled_sum_sq_diffs,
        sum_sq_diffs,
        prod_sq_diffs,
        sum_quad_diffs,
        a,
        length_scale,
    )
    full_m[..., 0::3, 1::3] = full_m[..., 1::3, 0::3] = _kg1_fn(
        exp_inv_scaled_sum_sq_diffs,
        diff_xy_quad_diffs,
        diff_yx_sq_diffs,
        a,
        length_scale,
    )
    full_m[..., 0::3, 2::3] = full_m[..., 2::3, 0::3] = _kg2_fn(
        exp_inv_scaled_sum_sq_diffs, sum_sq_diffs, prod_diffs, a, length_scale
    )
    full_m[..., 1::3, 1::3] = _g1g1_fn(
        exp_inv_scaled_sum_sq_diffs,
        sum_sq_diffs,
        sum_quad_diffs,
        prod_sq_diffs,
        a,
        length_scale,
    )
    full_m[..., 1::3, 2::3] = full_m[..., 2::3, 1::3] = _g1g2_fn(
        exp_inv_scaled_sum_sq_diffs,
        diff_xy_sq_diffs,
        prod_diffs,
        a,
        length_scale,
    )
    full_m[..., 2::3, 2::3] = _g2g2_fn(
        exp_inv_scaled_sum_sq_diffs,
        sum_sq_diffs,
        prod_sq_diffs,
        a,
        length_scale,
    )

    return full_m



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from typing import Union

import MuyGPyS._src.math.numpy as np
from MuyGPyS._src.gp.tensors import _pairwise_differences
from MuyGPyS.gp.kernels import Matern
from MuyGPyS.gp.noise import HeteroscedasticNoise, HomoscedasticNoise, NullNoise
from MuyGPyS.gp.sigma_sq import SigmaSq


class BenchmarkGP:
    """
    A basic Gaussian Process.

    Performs GP inference and simulation by way of analytic computations.

    Args:
        kernel:
            The kernel to be used. Only supports Matern.
        eps:
            The noise model.
    """

    def __init__(
        self,
        kernel: Matern,
        eps: Union[
            HeteroscedasticNoise, HomoscedasticNoise, NullNoise
        ] = HomoscedasticNoise(0.0),
    ):
        """
        Initialize.
        """
        self.kernel = kernel
        self.eps = eps
        self.sigma_sq = SigmaSq()

    def fixed(self) -> bool:
        """
        Checks whether all kernel and model parameters are fixed.

        This is a convenience utility to determine whether optimization is
        required.

        Returns:
            Returns `True` if all parameters are fixed, and `False` otherwise.
        """
        for p in self.kernel._hyperparameters:
            if not self.kernel._hyperparameters[p].fixed():
                return False
        if not self.eps.fixed():
            return False
        return True


def benchmark_sample_full(
    gp: BenchmarkGP,
    test: np.ndarray,
    train: np.ndarray,
) -> np.ndarray:
    """
    Sample from a GP prior for a dataset separated into train and test.

    Args:
        gp:
            The gp object
        test:
            The full testing data matrix of shape
            `(test_count, feature_count)`.
        train:
            The full training data matrix of shape
            `(train_count, feature_count)`.

    Returns:
        A sample from the GP prior for a train/test split.
    """
    return benchmark_sample(gp, np.vstack((test, train)))


def benchmark_prepare_cholK(
    gp: BenchmarkGP,
    data: np.ndarray,
) -> np.ndarray:
    """
    Sample from a GP prior for a dataset.

    Args:
        gp:
            The gp object
        train:
            The full training data matrix of shape
            `(train_count, feature_count)`.

    Returns:
        The Cholesky decomposition of a dense covariance matrix.
    """
    pairwise_diffs = _pairwise_differences(data)
    data_count, _ = data.shape
    Kfull = gp.sigma_sq()[0] * (
        gp.kernel(pairwise_diffs) + gp.eps() * np.eye(data_count)
    )
    return np.linalg.cholesky(Kfull)


def benchmark_sample(
    gp: BenchmarkGP,
    data: np.ndarray,
) -> np.ndarray:
    """
    Sample from a GP prior for a dataset.

    Args:
        gp:
            The gp object
        train:
            The full training data matrix of shape
            `(train_count, feature_count)`.
    """
    cholK = benchmark_prepare_cholK(gp, data)
    return benchmark_sample_from_cholK(cholK)


def benchmark_sample_from_cholK(cholK: np.ndarray) -> np.ndarray:
    data_count, _ = cholK.shape
    return (
        cholK @ np.array(np.random.normal(0, 1, size=(data_count,)))
    ).reshape(data_count, 1)


def get_analytic_sigma_sq(K, y):
    assert y.shape[0] == K.shape[0]
    return (1 / y.shape[0]) * y.T @ np.linalg.solve(K, y)



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

"""
MuyGPs implementation
"""

from typing import Callable, Union

import MuyGPyS._src.math as mm
from MuyGPyS._src.gp.muygps import _muygps_posterior_mean
from MuyGPyS.gp.noise import (
    HeteroscedasticNoise,
    HomoscedasticNoise,
    NullNoise,
    perturb_with_noise_model,
)


class PosteriorMean:
    def __init__(
        self,
        eps: Union[HeteroscedasticNoise, HomoscedasticNoise, NullNoise],
        **kwargs
    ):
        self.eps = eps
        self._fn = _muygps_posterior_mean
        self._fn = perturb_with_noise_model(self._fn, self.eps)

    def __call__(
        self,
        K: mm.ndarray,
        Kcross: mm.ndarray,
        batch_nn_targets: mm.ndarray,
    ) -> mm.ndarray:
        return self._fn(K, Kcross, batch_nn_targets, eps=self.eps())

    def get_opt_fn(self) -> Callable:
        return self._get_opt_fn(self._fn, self.eps)

    @staticmethod
    def _get_opt_fn(
        mean_fn: Callable,
        eps: Union[HeteroscedasticNoise, HomoscedasticNoise, NullNoise],
    ) -> Callable:
        opt_fn = eps.apply(mean_fn, "eps")
        return opt_fn



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

"""
Convenience functions for optimizing :class:`~MuyGPyS.gp.muygps.MuyGPS` objects

The functions
:func:`~MuyGPyS.optimize.chassis.optimize_from_indices` and
:func:`~MuyGPyS.optimize.chassis.optimize_from_tensors` wrap different
optimization packages to provide a simple interface to optimize the
hyperparameters of :class:`~MuyGPyS.gp.muygps.MuyGPS` objects.

Currently, `opt_method="scipy"` wraps :class:`scipy.optimize.opt`
multiparameter optimization using L-BFGS-B algorithm using the objective
function :func:`MuyGPyS.optimize.objective.loo_crossval`.

Currently, `opt_method="bayesian"` (also accepts `"bayes"` and `"bayes_opt"`)
wraps :class:`bayes_opt.BayesianOptimization`. Unlike the `scipy` version,
`BayesianOptimization` can be meaningfully modified by several kwargs.
`MuyGPyS` assigns reasonable defaults if no settings are passed by the user.
See the `BayesianOptimization <https://github.com/fmfn/BayesianOptimization>`_
documentation for details.
"""

from copy import deepcopy
import numpy as np
from time import process_time
from typing import Callable, Dict, Optional, Tuple

from bayes_opt import BayesianOptimization
import MuyGPyS._src.math as mm
from MuyGPyS._src.optimize.chassis.numpy import (
    _new_muygps,
    _get_opt_lists,
    _bayes_get_kwargs,
)
from MuyGPyS.gp import MuyGPS
from MuyGPyS.gp.distortion import AnisotropicDistortion
from MuyGPyS.gp.tensors import make_train_tensors
from MuyGPyS.neighbors import NN_Wrapper
from MuyGPyS.optimize.objective import make_obj_fn
from MuyGPyS.optimize.loss import lool_fn
from MuyGPyS.optimize.sigma_sq import (
    make_sigma_sq_optim,
    muygps_sigma_sq_optim,
)


def optimize_from_tensors_mini_batch(
    muygps: MuyGPS,
    train_features: mm.ndarray,
    train_responses: mm.ndarray,
    nn_count: int,
    batch_count: int,
    train_count: int,
    num_epochs: int = 1,
    keep_state: bool = False,
    probe_previous: bool = False,
    batch_features: Optional[mm.ndarray] = None,
    loss_fn: Callable = lool_fn,
    obj_method: str = "loo_crossval",
    sigma_method: Optional[str] = "analytic",
    loss_kwargs: Optional[Dict] = dict(),
    verbose: bool = False,
    **kwargs,
) -> Tuple[MuyGPS, NN_Wrapper, float, int, int]:
    """
    Find the optimal model using:
    1. Exact KNN using scikit learn
    2. Bayes Optimization
    3. numpy math backend

    See the following example, where we have already initialized a
    :class:`~MuyGPyS.gp.muygps.MuyGPS` model `muygps` and created a
    :class:`utils.UnivariateSampler` or :class:`utils.UnivariateSampler2D`
    instance `sampler`.

    Example:
        >>> (
        >>>     muygps_optloop,
        >>>     nbrs_lookup_final,
        >>>     exec_time,
        >>>     probe_count,
        >>>     opt_steps,
        >>> ) = optimize_from_tensors_mini_batch(
        ...     muygps,
        ...     train_features,
        ...     train_responses,
        ...     nn_count=30,
        ...     batch_count=sampler.train_count,
        ...     train_count=sampler.train_count,
        ...     num_epochs=1,
        ...     keep_state=False,
        ...     probe_previous=False,
        ...     loss_fn=lool_fn,
        ...     obj_method="loo_crossval",
        ...     verbose=True,
        ...     random_state=1,
        ...     init_points=5,
        ...     n_iter=20,
        ...     allow_duplicate_points=True,
        ... )
        parameters to be optimized: ['length_scale0', 'length_scale1']
        bounds: [[0.01 1.  ]
        [0.01 1.  ]]
        initial x0: [0.09718538 0.42218699]
        |   iter    |  target   | length... | length... |
        -------------------------------------------------
        | 1         | 793.7     | 0.09719   | 0.4222    |
        | 2         | 425.1     | 0.4229    | 0.7231    |
        | 3         | 39.33     | 0.01011   | 0.3093    |
        | 4         | 83.32     | 0.1553    | 0.1014    |
        | 5         | 480.6     | 0.1944    | 0.3521    |
        | 6         | 277.1     | 0.4028    | 0.5434    |
        | 7         | 790.1     | 0.09769   | 0.4123    |
        | 8         | 685.2     | 0.1737    | 0.4721    |
        | 9         | 792.9     | 0.09583   | 0.4147    |
        | 10        | 513.9     | 0.03333   | 0.6004    |
        | 11        | -444.6    | 1.0       | 0.01      |
        | 12        | 803.1     | 0.187     | 0.9781    |
        | 13        | -181.9    | 0.01      | 1.0       |
        | 14        | 707.7     | 0.3298    | 0.9882    |
        | 15        | 234.0     | 0.5577    | 0.705     |
        | 16        | 675.0     | 0.157     | 0.4126    |
        | 17        | 755.4     | 0.2337    | 0.8346    |
        | 18        | 107.2     | 1.0       | 1.0       |
        | 19        | -108.3    | 1.0       | 0.5391    |
        | 20        | -205.5    | 0.5192    | 0.01      |
        | 21        | 411.2     | 0.5939    | 1.0       |
        | 22        | 729.1     | 0.2109    | 0.6687    |
        | 23        | -98.07    | 0.673     | 0.3635    |
        | 24        | 797.6     | 0.1144    | 0.775     |
        | 25        | 761.3     | 0.2273    | 0.8345    |
        | 26        | 147.3     | 0.7809    | 0.845     |
        =================================================

        0, {'length_scale0': 0.1869..., 'length_scale1': 0.9781...}

    Args:
        muygps:
            The model to be optimized.
        train_features:
            Explanatory variables used to train model.
        train_responses:
            Labels corresponding to features used to train model.
        nn_count:
            The number of nearest neighbors to return in queries.
        batch_count:
            The number of batch elements to sample.
        train_count:
            The total number of training examples.
        num_epochs:
            The number of iterations for optimization loop.
        keep_state:
            If True, maintain optimizer target space and explored points.
        probe_previous:
            If True, store max params to probe in next loop iteration.
        batch_features:
            Set to None, ignore hierarchical stuff for now.
        loss_fn:
            The loss functor used to evaluate model performance.
        obj_method:
            Indicates the objective function to be minimized. Currently
            restricted to `"loo_crossval"`.
        sigma_method:
            The optimization method to be employed to learn the `sigma_sq`
            hyperparameter.
        loss_kwargs:
            A dictionary of additional keyword arguments to apply to the loss
            function. Loss function specific.
        verbose:
            If True, print debug messages.
        kwargs:
            Additional keyword arguments to be passed to the wrapper
            optimizer.

    Returns:
        A new MuyGPs model whose specified hyperparameters have been optimized.

    Returns
    -------
        MuyGPS:
            A new MuyGPs model with optimized hyperparameters.
        NN_Wrapper:
            Trained nearest neighbor query data structure.
        float:
            Total cpu and system time in loop execution.
        int:
            Total points probed (exploration).
        int:
            Total iterations of bayes optimization (exploitation).
    """

    # Get objective function components
    kernel_fn = muygps.kernel.get_opt_fn()
    mean_fn = muygps.get_opt_mean_fn()
    var_fn = muygps.get_opt_var_fn()
    sigma_sq_fn = make_sigma_sq_optim(sigma_method, muygps)

    # Create bayes_opt kwargs
    x0_names, x0, bounds = _get_opt_lists(muygps, verbose=verbose)
    x0_map = {n: x0[i] for i, n in enumerate(x0_names)}
    to_probe = [x0_map]
    bounds_map = {n: bounds[i] for i, n in enumerate(x0_names)}
    optimizer_kwargs, maximize_kwargs = _bayes_get_kwargs(
        verbose=verbose,
        **kwargs,
    )
    total_pts_probed = num_epochs * maximize_kwargs["init_points"]
    total_opt_steps = num_epochs * maximize_kwargs["n_iter"]

    # Create Bayes optimizer
    optimizer = BayesianOptimization(
        f=None,
        pbounds=bounds_map,
        **optimizer_kwargs,
    )
    optimized_values = ["\r\n"]

    # Initialize nearest neighbors lookup and get batch indices
    nbrs_lookup = NN_Wrapper(
        train_features, nn_count, nn_method="exact", algorithm="ball_tree"
    )
    new_nbrs_lookup = deepcopy(nbrs_lookup)
    batch_indices = mm.arange(train_count, dtype=mm.itype)

    # Run optimization loop
    time_start = process_time()
    for epoch in range(num_epochs):
        # Get batch nearest neighbors indices
        if not keep_state and train_count > batch_count:
            batch_indices = mm.iarray(
                np.random.choice(train_count, batch_count, replace=False)
            )
        batch_nn_indices, _ = new_nbrs_lookup.get_batch_nns(batch_indices)

        # Coalesce distance and target tensors
        (
            batch_crosswise_diffs,
            batch_pairwise_diffs,
            batch_targets,
            batch_nn_targets,
        ) = make_train_tensors(
            batch_indices,
            batch_nn_indices,
            train_features,
            train_responses,
        )

        # Generate the objective function
        obj_fn = make_obj_fn(
            obj_method,
            loss_fn,
            kernel_fn,
            mean_fn,
            var_fn,
            sigma_sq_fn,
            batch_pairwise_diffs,
            batch_crosswise_diffs,
            batch_nn_targets,
            batch_targets,
            batch_features=batch_features,
            loss_kwargs=loss_kwargs,
        )

        # Setting the function to be optimized for this epoch
        if keep_state:
            optimizer._space.target_func = obj_fn
        else:
            optimizer = BayesianOptimization(
                f=obj_fn,
                pbounds=bounds_map,
                **optimizer_kwargs,
            )

        # Probe explored points
        if probe_previous:
            for point in to_probe:
                optimizer.probe(point, lazy=True)
                total_pts_probed += 1
        elif epoch == 0:
            optimizer.probe(to_probe[0], lazy=True)
            total_pts_probed += 1

        # Find maximum of the acquisition function
        optimizer.maximize(**maximize_kwargs)

        # Add explored points to be probed
        to_probe.append(optimizer.max["params"])
        optimized_values.append(f"{epoch}, {optimizer.max['params']}")

        # Update neighborhoods using the learned length scales
        if isinstance(muygps.kernel.distortion_fn, AnisotropicDistortion) and (
            epoch < (num_epochs - 1)
        ):
            length_scales = AnisotropicDistortion._get_length_scale_array(
                mm.array,
                train_features.shape,
                None,
                **optimizer.max["params"],
            )
            train_features_scaled = train_features / length_scales
            new_nbrs_lookup = NN_Wrapper(
                train_features_scaled,
                nn_count,
                nn_method="exact",
                algorithm="ball_tree",
            )
    time_stop = process_time()

    # Print outcomes
    if verbose:
        for line in optimized_values:
            print(line)

    # Compute optimal variance scaling hyperparameter
    new_muygpys = muygps_sigma_sq_optim(
        _new_muygps(muygps, x0_names, bounds, optimizer.max["params"]),
        batch_pairwise_diffs,
        batch_nn_targets,
        sigma_method=sigma_method,
    )

    return (
        new_muygpys,
        nbrs_lookup,
        (time_stop - time_start),
        total_pts_probed,
        total_opt_steps,
    )



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

"""
RBF kernel functor

Defines RBF (or Gaussian, or squared exponential) kernel  (inheriting
:class:`~MuyGPyS.gp.kernels.kernel_fn.KernelFn`) that transform crosswise and
pairwise difference tensors into cross-covariance and covariance (or kernel)
tensors, respectively.

See the following example to initialize an :class:`MuyGPyS.gp.kernels.Matern`
object. Other kernel functors are similar, but require different
hyperparameters.

Example:
    >>> kernel_fn = RBF(
    ...     metric=IsotropicDistortion(
    ...         l2,
    ...         length_scale=ScalarHyperparameter(1.0),
    ...     ),
    ... )

One uses a previously computed `pairwise_diffs` tensor (see
:func:`MuyGPyS.gp.tensors.pairwise_tensor`) to compute a kernel tensor whose
second two dimensions contain square kernel matrices. Similarly, one uses a
previously computed `crosswise_diffs` matrix (see
:func:`MuyGPyS.gp.tensors.crosswise_tensor`) to compute a cross-covariance
matrix. See the following example, which assumes that you have already
constructed the difference `numpy.nparrays` and the kernel `kernel_fn` as shown
above.

Example:
    >>> K = kernel_fn(pairwise_diffs)
    >>> Kcross = kernel_fn(crosswise_diffs)
"""

from typing import Callable, List, Tuple, Union

import MuyGPyS._src.math as mm
from MuyGPyS._src.gp.kernels import _rbf_fn
from MuyGPyS._src.util import auto_str
from MuyGPyS.gp.distortion import (
    embed_with_distortion_model,
    AnisotropicDistortion,
    IsotropicDistortion,
    F2,
)
from MuyGPyS.gp.kernels import KernelFn
from MuyGPyS.gp.hyperparameter import ScalarHyperparameter


@auto_str
class RBF(KernelFn):
    """
    The radial basis function (RBF) or squared-exponential kernel.

    The RBF kernel includes a parameterized scaled distance function
    :math:`d_\\ell(\\cdot, \\cdot)`.

    The kernel is defined by

    .. math::
        K(x_i, x_j) = \\exp\\left(- d_\\ell(x_i, x_j)\\right).

    Typically, :math:`d(\\cdot,\\cdot)` is the squared Euclidean distance
    or second frequency moment of the difference of the operands.

    Args:
        metric:
            The distance function to be used. Includes length_scale
            hyperparameter information via the MuyGPyS.gp.distortion module
    """

    def __init__(
        self,
        metric: Union[
            AnisotropicDistortion, IsotropicDistortion
        ] = IsotropicDistortion(F2, length_scale=ScalarHyperparameter(1.0)),
    ):
        super().__init__(metric=metric)
        self._kernel_fn = _rbf_fn
        self._make()

    def _make(self):
        super()._make_base()
        self._fn = embed_with_distortion_model(
            self._kernel_fn,
            self.distortion_fn,
            self.distortion_fn.length_scale,
        )

    def __call__(
        self, diffs: mm.ndarray, batch_features: mm.ndarray = None, **kwargs
    ) -> mm.ndarray:
        """
        Compute RBF kernel(s) from a difference tensor.

        Args:
            diffs:
                A tensor of pairwise diffs of shape
                `(data_count, nn_count, nn_count, feature_count)` or
                `(data_count, nn_count, feature_count)`. In the four dimensional
                case, it is assumed that the diagonals dists
                diffs[i, j, j, :] == 0.
            batch_features:
                A tensor of shape `(data_count, 1)` or a vector of length
                `data_count` or a scalar.

        Returns:
            A cross-covariance matrix of shape `(data_count, nn_count)` or a
            tensor of shape `(data_count, nn_count, nn_count)` whose last two
            dimensions are kernel matrices.
        """
        return self._fn(diffs, batch_features=batch_features, **kwargs)

    def get_opt_params(
        self,
    ) -> Tuple[List[str], List[float], List[Tuple[float, float]]]:
        """
        Report lists of unfixed hyperparameter names, values, and bounds.

        Returns
        -------
            names:
                A list of unfixed hyperparameter names.
            params:
                A list of unfixed hyperparameter values.
            bounds:
                A list of unfixed hyperparameter bound tuples.
        """
        names, params, bounds = super().get_opt_params()
        return names, params, bounds

    def get_opt_fn(self) -> Callable:
        """
        Return a kernel function with fixed parameters set.

        This function is designed for use with
        :func:`MuyGPyS.optimize.chassis.optimize_from_tensors()` and assumes
        that optimization parameters will be passed as keyword arguments.

        Returns:
            A function implementing the kernel where all fixed parameters are
            set. The function expects keyword arguments corresponding to current
            hyperparameter values for unfixed parameters.
        """
        return super()._get_opt_fn(self._fn, self.distortion_fn)



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from mpi4py import MPI

import MuyGPyS._src.math.numpy as np
from MuyGPyS import config
from MuyGPyS._src.optimize.loss.numpy import (
    _mse_fn_unnormalized,
    _cross_entropy_fn as _cross_entropy_fn_n,
    _lool_fn as _lool_fn_n,
    _pseudo_huber_fn as _pseudo_huber_fn_n,
    _looph_fn as _looph_fn_n,
)

world = config.mpi_state.comm_world


def _mse_fn(
    predictions: np.ndarray,
    targets: np.ndarray,
) -> float:
    local_batch_count, response_count = predictions.shape
    local_squared_errors = _mse_fn_unnormalized(predictions, targets)
    global_batch_count = world.allreduce(local_batch_count, op=MPI.SUM)
    global_squared_errors = world.allreduce(local_squared_errors, op=MPI.SUM)
    return global_squared_errors / (global_batch_count * response_count)


def _make_mpi_obj_fn(kwargs_opt_fn, comm):
    def caller_fn(**kwargs):
        local_obj = kwargs_opt_fn(**kwargs)
        global_obj = comm.allreduce(local_obj, op=MPI.SUM)
        return global_obj

    return caller_fn


def _cross_entropy_fn(
    predictions: np.ndarray,
    targets: np.ndarray,
    ll_eps: float = 1e-15,
) -> float:
    local_log_loss = _cross_entropy_fn_n(predictions, targets, ll_eps=ll_eps)
    global_log_loss = world.allreduce(local_log_loss, op=MPI.SUM)
    return global_log_loss


def _lool_fn(
    predictions: np.ndarray,
    targets: np.ndarray,
    variances: np.ndarray,
    sigma_sq: np.ndarray,
) -> float:
    local_likelihoods = _lool_fn_n(predictions, targets, variances, sigma_sq)
    global_likelihood = world.allreduce(local_likelihoods, op=MPI.SUM)
    return global_likelihood


def _lool_fn_unscaled(
    predictions: np.ndarray, targets: np.ndarray, variances: np.ndarray
) -> float:
    local_likelihoods = _lool_fn_unscaled(predictions, targets, variances)
    global_likelihood = world.allreduce(local_likelihoods, op=MPI.SUM)
    return global_likelihood


def _pseudo_huber_fn(
    predictions: np.ndarray, targets: np.ndarray, boundary_scale: float = 1.5
) -> float:
    local_pseudo_huber = _pseudo_huber_fn_n(
        predictions, targets, boundary_scale=boundary_scale
    )
    global_pseudo_huber = world.allreduce(local_pseudo_huber, op=MPI.SUM)
    return global_pseudo_huber


def _looph_fn(
    predictions: np.ndarray,
    targets: np.ndarray,
    variances: np.ndarray,
    sigma_sq: np.ndarray,
    boundary_scale: float = 1.5,
) -> float:
    local_looph = _looph_fn_n(
        predictions, targets, variances, sigma_sq, boundary_scale=boundary_scale
    )
    global_looph = world.allreduce(local_looph, op=MPI.SUM)
    return global_looph



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from MuyGPyS._src.util import _collect_implementation

(
    _make_predict_tensors,
    _make_fast_predict_tensors,
    _make_train_tensors,
    _batch_features_tensor,
    _crosswise_differences,
    _crosswise_tensor,
    _pairwise_differences,
    _pairwise_tensor,
    _fast_nn_update,
    _make_heteroscedastic_tensor,
    _F2,
    _l2,
) = _collect_implementation(
    "MuyGPyS._src.gp.tensors",
    "_make_predict_tensors",
    "_make_fast_predict_tensors",
    "_make_train_tensors",
    "_batch_features_tensor",
    "_crosswise_differences",
    "_crosswise_tensor",
    "_pairwise_differences",
    "_pairwise_tensor",
    "_fast_nn_update",
    "_make_heteroscedastic_tensor",
    "_F2",
    "_l2",
)



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

"""
MuyGPs implementation
"""

from typing import Union

import MuyGPyS._src.math as mm
from MuyGPyS._src.gp.muygps import (
    _muygps_fast_posterior_mean_precompute,
)
from MuyGPyS.gp.noise import (
    HomoscedasticNoise,
    HeteroscedasticNoise,
    NullNoise,
    perturb_with_noise_model,
)


class FastPrecomputeCoefficients:
    def __init__(
        self,
        eps: Union[HeteroscedasticNoise, HomoscedasticNoise, NullNoise],
        **kwargs
    ):
        self.eps = eps
        self._fn = _muygps_fast_posterior_mean_precompute
        self._fn = perturb_with_noise_model(self._fn, self.eps)

    def __call__(
        self,
        K: mm.ndarray,
        train_nn_targets_fast: mm.ndarray,
    ) -> mm.ndarray:
        return self._fn(K, train_nn_targets_fast, eps=self.eps())
